/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 * 
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tud.kom.p2psim.impl.topology.views.fiveg;

import java.util.HashMap;

import de.tud.kom.p2psim.api.linklayer.mac.MacAddress;
import de.tud.kom.p2psim.api.linklayer.mac.PhyType;
import de.tud.kom.p2psim.api.scenario.ConfigurationException;
import de.tud.kom.p2psim.impl.linklayer.ModularLinkLayer;
import de.tud.kom.p2psim.impl.topology.views.fiveg.models.AbstractModel;
import de.tud.kom.p2psim.impl.topology.views.fiveg.models.ParameterType;
import de.tud.kom.p2psim.impl.util.IncrementableHashMap;
import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.LocationListener;

public class ModelBasedSegmentDatabase extends AbstractGridBasedTopologyDatabase
		implements LocationListener {
	
	final Integer DEFAULT_SEGMENT_ID = new Integer(-1);

	private HashMap<MacAddress, Integer> currentPositions = new HashMap<>();

	IncrementableHashMap<Integer> usersPerSegment = new IncrementableHashMap<>();

	HashMap<ParameterType, HashMap<Integer, AbstractModel>> models = new HashMap<>();

	public ModelBasedSegmentDatabase() {
		super(100, true);
		super.setSupportCloudlets(true);
	}

	@Override
	protected Entry createEntryFor(int segmentID, boolean isCloudlet) {
		// TODO Handle Cloudlets differently?
		return new ModelBasedEntry(segmentID);
	}

	public void setModel(AbstractModel model) {
		ParameterType type = model.getParameterType();
		if (type == null) {
			throw new ConfigurationException(
					"Model is not defined for any Parameter Type. Please specify the Type for this Model.");
		}

		Integer segmentID = new Integer(model.getSegmentID());

		if (models.containsKey(type)) {
			if (models.get(type).containsKey(segmentID)) {
				throw new ConfigurationException(
						"Conflicting models for " + type.toString()
								+ " and SegementID " + segmentID.toString());
			} else {
				models.get(type).put(segmentID, model);
			}
		} else {
			HashMap<Integer, AbstractModel> tmp = new HashMap<>();
			tmp.put(segmentID, model);
			models.put(type, tmp);
		}
	}

	@Override
	public void onLocationChanged(Host host, Location location) {
		try {
			// Get MAC Address and the SegementID the user is now in
			MacAddress mac = host.getComponent(ModularLinkLayer.class)
					.getMac(PhyType.UMTS).getMacAddress();
			Integer segmentID = new Integer(this.getSegmentID(
					location.getLongitude(), location.getLatitude()));
			// Check if it is a new host
			if (currentPositions.containsKey(mac)) {
				Integer oldSegmentID = currentPositions.get(mac);
				// Host changed segment
				if (oldSegmentID.compareTo(segmentID) != 0) {
					usersPerSegment.decrement(oldSegmentID);
					usersPerSegment.increment(segmentID);
					currentPositions.put(mac, segmentID);
				}
			} else {
				// New host
				currentPositions.put(mac, segmentID);
				usersPerSegment.increment(segmentID);
			}
		} catch (ComponentNotAvailableException e) {
			// TODO Handle unknown host type
		}
	}

	public class ModelBasedEntry implements FiveGTopologyDatabase.Entry {

		private final int segment;

		private boolean connectivity = true;

		public ModelBasedEntry(int segment) {
			this.segment = segment;
		}

		@Override
		public int getSegmentID() {
			return segment;
		}

		@Override
		public double getDropProbability(boolean isUpload) {
			if (!getConnectivity()) {
				return 1;
			}
			// TODO Up / Download?
			Integer segmentID = new Integer(getSegmentID());
			int usersInSegment = usersPerSegment.get(segmentID);
			
			// Is there a Model for Droprate?
			if (models.containsKey(ParameterType.DROPRATE)) {
				// Is there a specific model for this segment
				if(models.get(ParameterType.DROPRATE).containsKey(segmentID)) {
					return models.get(ParameterType.DROPRATE).get(segmentID).getDouble(usersInSegment);
				} else {
					return models.get(ParameterType.DROPRATE).get(DEFAULT_SEGMENT_ID).getDouble(usersInSegment);
				}
			} else {
				throw new ConfigurationException(
						"No Model is defined for Droprate.");
			}

		}

		@Override
		public long getLatency(boolean isUpload) {
			if (!getConnectivity()) {
				return 9999 * Time.MILLISECOND;
			}
			// TODO Return Latency based on Model
			return 0;
		}

		@Override
		public long getBandwidth(boolean isUpload) {
			if (!getConnectivity()) {
				return 0;
			}
			// TODO Return Bandwidth based on Model
			return 0;
		}

		@Override
		public boolean getConnectivity() {
			return connectivity;
		}

		@Override
		public void setConnectivity(boolean connectivity) {
			this.connectivity = connectivity;
		}
	}
}
